@@ -88,42 +88,6 @@ typedef struct cairo_stroker {
     cairo_box_t bounds;
 } cairo_stroker_t;
 
-static cairo_status_t
-_cairo_stroker_init (cairo_stroker_t		*stroker,
-		     const cairo_stroke_style_t	*stroke_style,
-		     const cairo_matrix_t	*ctm,
-		     const cairo_matrix_t	*ctm_inverse,
-		     double			 tolerance)
-{
-    cairo_status_t status;
-
-    stroker->style = *stroke_style;
-    stroker->ctm = ctm;
-    stroker->ctm_inverse = ctm_inverse;
-    stroker->tolerance = tolerance;
-
-    stroker->ctm_determinant = _cairo_matrix_compute_determinant (stroker->ctm);
-    stroker->ctm_det_positive = stroker->ctm_determinant >= 0.0;
-
-    status = _cairo_pen_init (&stroker->pen,
-		              stroke_style->line_width / 2.0,
-			      tolerance, ctm);
-    if (unlikely (status))
-	return status;
-
-    stroker->has_bounds = FALSE;
-
-    stroker->has_current_face = FALSE;
-    stroker->has_first_face = FALSE;
-    stroker->has_initial_sub_path = FALSE;
-
-    _cairo_stroker_dash_init (&stroker->dash, stroke_style);
-
-    stroker->add_external_edge = NULL;
-
-    return CAIRO_STATUS_SUCCESS;
-}
-
 static void
 _cairo_stroker_limit (cairo_stroker_t *stroker,
 		      const cairo_path_fixed_t *path,
@@ -154,6 +118,47 @@ _cairo_stroker_limit (cairo_stroker_t *stroker,
     stroker->bounds.p2.y += fdy;
 }
 
+static cairo_status_t
+_cairo_stroker_init (cairo_stroker_t		*stroker,
+		     const cairo_path_fixed_t	*path,
+		     const cairo_stroke_style_t	*stroke_style,
+		     const cairo_matrix_t	*ctm,
+		     const cairo_matrix_t	*ctm_inverse,
+		     double			 tolerance,
+		     const cairo_box_t		*limits,
+		     int			 num_limits)
+{
+    cairo_status_t status;
+
+    stroker->style = *stroke_style;
+    stroker->ctm = ctm;
+    stroker->ctm_inverse = ctm_inverse;
+    stroker->tolerance = tolerance;
+
+    stroker->ctm_determinant = _cairo_matrix_compute_determinant (stroker->ctm);
+    stroker->ctm_det_positive = stroker->ctm_determinant >= 0.0;
+
+    status = _cairo_pen_init (&stroker->pen,
+		              stroke_style->line_width / 2.0,
+			      tolerance, ctm);
+    if (unlikely (status))
+	return status;
+
+    stroker->has_current_face = FALSE;
+    stroker->has_first_face = FALSE;
+    stroker->has_initial_sub_path = FALSE;
+
+    _cairo_stroker_dash_init (&stroker->dash, stroke_style);
+
+    stroker->add_external_edge = NULL;
+
+    stroker->has_bounds = FALSE;
+    if (num_limits)
+	_cairo_stroker_limit (stroker, path, limits, num_limits);
+
+    return CAIRO_STATUS_SUCCESS;
+}
+
 static void
 _cairo_stroker_fini (cairo_stroker_t *stroker)
 {
@@ -220,100 +225,117 @@ _tessellate_fan (cairo_stroker_t *stroker,
 		 cairo_bool_t clockwise)
 {
     cairo_point_t stack_points[64], *points = stack_points;
-    int start, stop, step, i, npoints;
+    cairo_pen_t *pen = &stroker->pen;
+    int start, stop, num_points = 0;
     cairo_status_t status;
 
-    if (clockwise) {
-	step  = -1;
-
-	start = _cairo_pen_find_active_ccw_vertex_index (&stroker->pen,
-							 in_vector);
-	if (_cairo_slope_compare (&stroker->pen.vertices[start].slope_ccw,
-				  in_vector) < 0)
-	    start = _range_step (start, -1, stroker->pen.num_vertices);
-
-	stop  = _cairo_pen_find_active_ccw_vertex_index (&stroker->pen,
-							 out_vector);
-	if (_cairo_slope_compare (&stroker->pen.vertices[stop].slope_cw,
-				  out_vector) > 0)
-	{
-	    stop = _range_step (stop, 1, stroker->pen.num_vertices);
-	    if (_cairo_slope_compare (&stroker->pen.vertices[stop].slope_ccw,
-				      in_vector) < 0)
-	    {
-		goto BEVEL;
-	    }
-	}
-
-	npoints = start - stop;
-    } else {
-	step  = 1;
-
-	start = _cairo_pen_find_active_cw_vertex_index (&stroker->pen,
-							in_vector);
-	if (_cairo_slope_compare (&stroker->pen.vertices[start].slope_cw,
-				  in_vector) < 0)
-	    start = _range_step (start, 1, stroker->pen.num_vertices);
-
-	stop  = _cairo_pen_find_active_cw_vertex_index (&stroker->pen,
-							out_vector);
-	if (_cairo_slope_compare (&stroker->pen.vertices[stop].slope_ccw,
-				  out_vector) > 0)
-	{
-	    stop = _range_step (stop, -1, stroker->pen.num_vertices);
-	    if (_cairo_slope_compare (&stroker->pen.vertices[stop].slope_cw,
-				      in_vector) < 0)
-	    {
-		goto BEVEL;
-	    }
-	}
-
-	npoints = stop - start;
-    }
-    stop = _range_step (stop, step, stroker->pen.num_vertices);
-
-    if (npoints < 0)
-	npoints += stroker->pen.num_vertices;
-    npoints += 3;
-
-    if (npoints <= 1)
+    if (stroker->has_bounds &&
+	! _cairo_box_contains_point (&stroker->bounds, midpt))
 	goto BEVEL;
 
-    if (npoints > ARRAY_LENGTH (stack_points)) {
-	points = _cairo_malloc_ab (npoints, sizeof (cairo_point_t));
-	if (unlikely (points == NULL))
-	    return _cairo_error (CAIRO_STATUS_NO_MEMORY);
-    }
+    assert (stroker->pen.num_vertices);
 
+    if (clockwise) {
+	_cairo_pen_find_active_ccw_vertices (pen,
+					     in_vector, out_vector,
+					     &start, &stop);
+	if (stroker->add_external_edge) {
+	    cairo_point_t last;
+	    last = *inpt;
+	    while (start != stop) {
+		cairo_point_t p = *midpt;
+		_translate_point (&p, &pen->vertices[start].point);
 
-    /* Construct the fan. */
-    npoints = 0;
-    points[npoints++] = *inpt;
-    for (i = start;
-	 i != stop;
-	i = _range_step (i, step, stroker->pen.num_vertices))
-    {
-	points[npoints] = *midpt;
-	_translate_point (&points[npoints], &stroker->pen.vertices[i].point);
-	npoints++;
-    }
-    points[npoints++] = *outpt;
-
-    if (stroker->add_external_edge != NULL) {
-	for (i = 0; i < npoints - 1; i++) {
-	    if (clockwise) {
 		status = stroker->add_external_edge (stroker->closure,
-						     &points[i], &points[i+1]);
-	    } else {
-		status = stroker->add_external_edge (stroker->closure,
-						     &points[i+1], &points[i]);
+						     &last, &p);
+		if (unlikely (status))
+		    return status;
+		last = p;
+
+		if (start-- == 0)
+		    start += pen->num_vertices;
 	    }
-	    if (unlikely (status))
-		break;
+	    status = stroker->add_external_edge (stroker->closure,
+						 &last, outpt);
+	} else {
+	    if (start == stop)
+		goto BEVEL;
+
+	    num_points = stop - start;
+	    if (num_points < 0)
+		num_points += pen->num_vertices;
+	    num_points += 2;
+	    if (num_points > ARRAY_LENGTH(stack_points)) {
+		points = _cairo_malloc_ab (num_points, sizeof (cairo_point_t));
+		if (unlikely (points == NULL))
+		    return _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	    }
+
+	    points[0] = *inpt;
+	    num_points = 1;
+	    while (start != stop) {
+		points[num_points] = *midpt;
+		_translate_point (&points[num_points], &pen->vertices[start].point);
+		num_points++;
+
+		if (start-- == 0)
+		    start += pen->num_vertices;
+	    }
+	    points[num_points++] = *outpt;
 	}
     } else {
+	_cairo_pen_find_active_cw_vertices (pen,
+					    in_vector, out_vector,
+					    &start, &stop);
+	if (stroker->add_external_edge) {
+	    cairo_point_t last;
+	    last = *inpt;
+	    while (start != stop) {
+		cairo_point_t p = *midpt;
+		_translate_point (&p, &pen->vertices[start].point);
+
+		status = stroker->add_external_edge (stroker->closure,
+						     &p, &last);
+		if (unlikely (status))
+		    return status;
+		last = p;
+
+		if (++start == pen->num_vertices)
+		    start = 0;
+	    }
+	    status = stroker->add_external_edge (stroker->closure,
+						 outpt, &last);
+	} else {
+	    if (start == stop)
+		goto BEVEL;
+
+	    num_points = stop - start;
+	    if (num_points < 0)
+		num_points += pen->num_vertices;
+	    num_points += 2;
+	    if (num_points > ARRAY_LENGTH(stack_points)) {
+		points = _cairo_malloc_ab (num_points, sizeof (cairo_point_t));
+		if (unlikely (points == NULL))
+		    return _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	    }
+
+	    points[0] = *inpt;
+	    num_points = 1;
+	    while (start != stop) {
+		points[num_points] = *midpt;
+		_translate_point (&points[num_points], &pen->vertices[start].point);
+		num_points++;
+
+		if (++start == pen->num_vertices)
+		    start = 0;
+	    }
+	    points[num_points++] = *outpt;
+	}
+    }
+
+    if (num_points) {
 	status = stroker->add_triangle_fan (stroker->closure,
-					    midpt, points, npoints);
+					    midpt, points, num_points);
     }
 
     if (points != stack_points)
@@ -1244,8 +1266,9 @@ _cairo_path_fixed_stroke_to_shaper (cairo_path_fixed_t	*path,
     cairo_stroker_t stroker;
     cairo_status_t status;
 
-    status = _cairo_stroker_init (&stroker, stroke_style,
-			          ctm, ctm_inverse, tolerance);
+    status = _cairo_stroker_init (&stroker, path, stroke_style,
+			          ctm, ctm_inverse, tolerance,
+				  NULL, 0);
     if (unlikely (status))
 	return status;
 
@@ -1286,18 +1309,15 @@ _cairo_path_fixed_stroke_dashed_to_polygon (const cairo_path_fixed_t	*path,
     cairo_stroker_t stroker;
     cairo_status_t status;
 
-    status = _cairo_stroker_init (&stroker, stroke_style,
-			          ctm, ctm_inverse, tolerance);
+    status = _cairo_stroker_init (&stroker, path, stroke_style,
+			          ctm, ctm_inverse, tolerance,
+				  polygon->limits, polygon->num_limits);
     if (unlikely (status))
 	return status;
 
     stroker.add_external_edge = _cairo_polygon_add_external_edge,
     stroker.closure = polygon;
 
-    if (polygon->num_limits)
-	_cairo_stroker_limit (&stroker, path,
-			      polygon->limits, polygon->num_limits);
-
     status = _cairo_path_fixed_interpret (path,
 					  _cairo_stroker_move_to,
 					  stroker.dash.dashed ?
